/*-
 * Copyright (C) @BABOLO  2008 Nov 26 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

 * One More Cisco PassWord Cryptor and DeCryptor inspired by
 * Id: ios7decrypt.pl,v 1.1 1998/01/11 21:31:12 mesrik Exp
 *
 * Credits for orginal code and description hobbit@avian.org,
 * SPHiXe, .mudge et al. and for John Bashinski <jbash@CISCO.COM>
 * for Cisco IOS password encryption facts.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2008 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: omcpwcdc.c,v 1.2 2008/11/27 20:19:42 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <fcntl.h>
#include <err.h>
#include <babolo/BLINflag.h>

#define OMCPWCDC_LIMITERCH 0x0000FF
#define OMCPWCDC_EVENODITY 0x000100
#define OMCPWCDC_DIRECTION 0x000200
#define OMCPWCDC_SALTDEFIN 0x000400
#define OMCPWCDC_ENABLIMIT 0x000800
#define OMCPWCDC_DISABLENL 0x001000

static u_char xlat[] = "dsfd;kfoA,.iyewrkldJKDHSUBsgvca69834ncxv9873254k;fg87dsf";

#define C_er 0x80000000 /*     -                                         */
#define C_d0 0x40000000 /*                                                 */
#define C_x0 0x20000000 /*                                                 */
#define C_X0 0x10000000 /*                                                 */
#define C_d1 0x08000000 /*                                                 */
#define C_x1 0x04000000 /*                                                 */
#define C_X1 0x02000000 /*                                                 */
#define C_n1 0x01000000 /*                                               */
#define C_rs 0x00800000 /*    salt                        */
#define C_rp 0x00400000 /*                                                     */
#define C_re 0x00200000 /*                                                    */
#define C_state    0xFF /*                                                        */

enum states {s0, s1, d0, d1, ww, dx};
/* s0 -  0  salt
   s1 -  1  salt
   d0 -   
   d1 -   
   ww - walk  salt
 */

enum classes {nl, no, ws, hd, hx, hX, xx};
/* nl \0
   no  
   ws white space
   hd  
   hx   
   hX   
   xx
 */

static const enum classes class[256] =
{nl,no,no,no ,no,no,no,no ,no,ws,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,ws,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,hd,hd,hd,hd ,hd,hd,hd,hd ,hd,hd,no,no ,no,no,no,no

,no,hX,hX,hX ,hX,hX,hX,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,hx,hx,hx ,hx,hx,hx,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no

,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no

,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
,no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
};

static const u_int32_t automa[xx][dx] =
/*          s0 ;          s1 ;          d0 ;          d1 ;          ww   */
{{          dx ,C_er|     dx ,C_re|     dx ,C_er|     dx ,          dx}/* \0 */
,{C_er|     s0 ,C_er|     dx ,C_er|     dx ,C_er|     dx ,C_er|     ww}/* - */
,{          s0 ,C_er|     dx ,C_er|     dx ,C_er|     dx ,          ww}/*   */
,{     C_d0|s1 ,C_rs|C_n1|d0 ,     C_d0|d1 ,C_rp|C_d1|d0 ,     C_d0|d1}/* 0 */
,{     C_x0|s1 ,C_er|     dx ,     C_x0|d1 ,C_rp|C_x1|d0 ,     C_x0|d1}/* x */
,{     C_X0|s1 ,C_er|     dx ,     C_X0|d1 ,C_rp|C_X1|d0 ,     C_X0|d1}/* X */
};

static void
usage(int ex) {
    fprintf( stderr
           , "omcpwcdc [-cqvv..][ -s salt] password\n"
             " -d      - decrypt instead of default encrypt\n"
             " -l char - use char as quotes\n"
             " -n      - no newline after password\n"
             " -s salt - first byte of encrypted word\n"
             " -q      - quiet\n"
             " -v..    - more verbose\n"
           );
    exit(ex);
}

int
main(int argc, char **argv, char **envp) {
    u_int32_t control = 0;
    u_int32_t flags = BLIN_VER1;
    u_char *p, *q, *r, value, salt = 0;
    enum states state;
    int c, ex = EX_OK, randf;

    while((c = getopt(argc, argv, "dl:ns:qv")) > 0) {
        switch(c) {
        case 'd': flags |= OMCPWCDC_DIRECTION;                      break;
        case 'l': flags |= ((*optarg) & 0xFF) | OMCPWCDC_ENABLIMIT; break;
        case 'n': flags |= OMCPWCDC_DISABLENL;                      break;
        case 's': salt = strtoul(optarg, NULL, 0);
                  flags |= OMCPWCDC_SALTDEFIN;                      break;
        case 'q': flags &= ~BLIN_VERMASK;                           break;
        case 'v': BLIN_VERBOSE(flags);                              break;
        default: usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc == 0) {
        ifBLIN_QV1(flags) warnx("Password missed");
        ex = EX_UNAVAILABLE;
    } else {
        if  (!argv || !*argv) {
            ifBLIN_QV1(flags) warnx("NULL (*)argv");
            ex = EX_USAGE;
        } else {
            p = *(u_char **)argv;
    }   }
    if  (!!ex) {
    } else if (flags & OMCPWCDC_DIRECTION) {
        r = q = p, value = 0;
        for (state = (flags & OMCPWCDC_SALTDEFIN) ? ww : s0; state < dx; state = control & C_state) {
            control = automa[class[*p]][state];
            if  (control & C_er) ex = -EX_DATAERR;
            if  (control & C_d0) value =   *(p++) - '0';
            if  (control & C_x0) value =   *(p++) - 'a' + 10;
            if  (control & C_X0) value =   *(p++) - 'A' + 10;
            if  (control & C_d1) value = ((*(p++) - '0'     ) & 0x0F) | (value << 4);
            if  (control & C_x1) value = ((*(p++) - 'a' + 10) & 0x0F) | (value << 4);
            if  (control & C_X1) value = ((*(p++) - 'a' + 10) & 0x0F) | (value << 4);
            if  (control & C_n1) value = ((*(p++) - '0'     ) & 0x0F) + (value * 10);
            if  (control & C_rs) salt = value, r = q = p;
            if  (control & C_rp) {
                *(q++) = value ^ xlat[salt++];
            }
            if  (control & C_re) {
                *q++ = '\0';
                if  (flags & OMCPWCDC_ENABLIMIT) printf("%c", flags & OMCPWCDC_LIMITERCH);
                printf("%s", r);
                if  (flags & OMCPWCDC_ENABLIMIT) printf("%c", flags & OMCPWCDC_LIMITERCH);
                if  (~flags & OMCPWCDC_DISABLENL) printf("\n");
            }
            if  (control & C_er) {
                ifBLIN_QV1(flags) warnx("input string syntax error");
                ex = EX_DATAERR;
            }
            while (salt > 55) salt -= 53;
        }
    } else {
        if  (~flags & OMCPWCDC_SALTDEFIN) {
            if  ((randf = open("/dev/random", O_RDONLY)) < 0) {
                ifBLIN_QV1(flags) warn("open /dev/random");
                ex = EX_IOERR;
            } else if (read(randf, &salt, 1) != 1) {
                ifBLIN_QV1(flags) warn("read /dev/random");
                ex = EX_IOERR;
            } else {
                close(randf);
        }   }
        if  (!ex) {
            if  (flags & OMCPWCDC_ENABLIMIT) printf("%c", flags & OMCPWCDC_LIMITERCH);
            printf("%02d", salt);
            for (r = q = p, c = 0; *p; p++) {
                while (salt > 55) salt -= 53;
                printf("%02X", *p ^ xlat[salt++]);
            }
            if  (flags & OMCPWCDC_ENABLIMIT) printf("%c", flags & OMCPWCDC_LIMITERCH);
            if  (~flags & OMCPWCDC_DISABLENL) printf("\n");
    }   }
    exit(ex);
}
